"""Module containing command-line argument parsing logic."""

import argparse
from argparse import ArgumentParser, RawTextHelpFormatter
from dataclasses import KW_ONLY, dataclass
from enum import StrEnum, auto
from pathlib import Path

from rust2rpm.sysinfo import TARGET_NAMES, Target, get_default_target


class VendorMode(StrEnum):
    """Description of different modes for using vendored dependencies."""

    OFF = auto()
    """Generate a spec file without any references to a vendor tarball."""

    AUTO = auto()
    """Generate a spec file for use with a vendor tarball, and create a
    tarball with vendored dependencies with default settings."""

    MANUAL = auto()
    """Generate a spec file for use with a vendor tarball, but require
    manual assembly of vendored sources into a tarball."""


VENDOR_MODES = [mode.value for mode in VendorMode]


def get_parser() -> ArgumentParser:
    """Construct an `argparse.ArgumentParser` for parsing command-line arguments."""
    parser = ArgumentParser("rust2rpm", formatter_class=RawTextHelpFormatter)

    # ==================== #
    # positional arguments #
    # ==================== #

    # pkgid: str | None = None
    parser.add_argument(
        "pkgid",
        nargs="?",
        help="""\
crate           (crate on crates.io)
crate@version   (pkgid on crates.io)
@version        (auto-detect crate name)""",
    )

    # ==================== #
    # options with a value #
    # ==================== #

    # config_file: str | None
    parser.add_argument(
        "-C",
        "--config-file",
        action="store",
        default=None,
        dest="config_file",
        help="Path to rust2rpm.toml config file (defaults to ./rust2rpm.toml)",
    )

    # output_directory: str | None
    parser.add_argument(
        "-o",
        "--output-directory",
        action="store",
        default=None,
        dest="output_directory",
        help="Target directory for any created files",
    )

    # path: str | None = None
    parser.add_argument(
        "--path",
        action="store",
        default=None,
        dest="path",
        help="Path to local project directory or Cargo.toml file to check instead of crates.io",
    )

    # target: str | None = None
    parser.add_argument(
        "-t",
        "--target",
        action="store",
        choices=TARGET_NAMES,
        default=None,
        dest="target",
        help="Distribution target",
    )

    # vendor: VendorMode = VendorMode.OFF
    parser.add_argument(
        "-V",
        "--vendor",
        action="store",
        choices=VENDOR_MODES,
        default=None,
        dest="vendor",
        help="Write a spec for building with vendored dependencies (default=off)",
    )

    # ==================== #
    # flag-only options    #
    # ==================== #

    # rpmautospec: bool | None = None
    parser.add_argument(
        "-a",
        "--rpmautospec",
        action="store_true",
        default=None,
        dest="rpmautospec",
        help="Use autorelease and autochangelog features",
    )
    parser.add_argument(
        "--no-rpmautospec",
        action="store_false",
        default=None,
        dest="rpmautospec",
        help="Do not use rpmautospec",
    )

    # auto_changelog_entry: bool = True
    parser.add_argument(
        "--no-auto-changelog-entry",
        action="store_false",
        dest="auto_changelog_entry",
        help=argparse.SUPPRESS,
    )

    # compat: bool = False
    parser.add_argument(
        "--compat",
        action="store_true",
        dest="compat",
        help="Create a compat package with appropriate version suffix",
    )

    # ignore_missing_license_files: bool = False
    parser.add_argument(
        "-I",
        "--ignore-missing-license-files",
        action="store_true",
        dest="ignore_missing_license_files",
        help=argparse.SUPPRESS,
    )

    # existence_check: bool = True
    parser.add_argument(
        "--no-existence-check",
        action="store_false",
        dest="existence_check",
        help="Do not check whether the package already exists in dist-git",
    )

    # offline: bool = False
    parser.add_argument(
        "-O",
        "--offline",
        action="store_true",
        dest="offline",
        help="Skip network queries to crates.io and / or src.fedoraproject.org",
    )

    # patch: bool = False
    parser.add_argument(
        "-p",
        "--patch",
        action="store_true",
        dest="patch",
        help="Do manual patching of Cargo.toml",
    )

    # patch_foreign: bool = True
    parser.add_argument(
        "--no-patch-foreign",
        action="store_false",
        dest="patch_foreign",
        help="Do not automatically drop foreign dependencies in Cargo.toml",
    )

    # relative_license_paths: bool = False
    parser.add_argument(
        "--relative-license-paths",
        action="store_true",
        dest="relative_license_paths",
        help=argparse.SUPPRESS,
    )

    # reuse_patch: bool = False
    parser.add_argument(
        "-r",
        "--reuse-patch",
        action="store_true",
        dest="reuse_patch",
        help="Attempt to re-apply existing Cargo.toml patch",
    )

    # store_crate: bool = False
    parser.add_argument(
        "-s",
        "--store-crate",
        action="store_true",
        dest="store_crate",
        help="Store crate in current directory",
    )

    # validate_only: bool = False
    parser.add_argument(
        "-v",
        "--validate-only",
        action="store_true",
        dest="validate_only",
        help="Validate rust2rpm.toml config file and exit",
    )

    return parser


@dataclass
class Options:
    """Collection of boolean or flag-like command-line arguments."""

    _: KW_ONLY
    # keep sorted alphabetically

    auto_changelog_entry: bool = True
    """Include an auto-generated changelog entry in rendered spec files
    (default: ON)."""

    compat: bool = False
    """Apply *"compat"* version suffis to the RPM `Name` and file name of the
    written spec file (default: OFF).
    """

    existence_check: bool = True
    """Check whether the package that a spec file is being generated for already
    exists in the distribution repositories (default: ON)."""

    ignore_missing_license_files: bool = False
    """Do not treat it as a hard error if the heuristics for detecting license
    files in project sources don't detect any files (default: OFF)."""

    offline: bool = False
    """Skip any operations that would require accessing the network and fail if
    any requested operation would require internet access default: OFF)."""

    patch: bool = False
    """Open an editor for interactively patching `Cargo.toml` metadata. This is
    the only supported way of editing `Cargo.toml` since it is the only way to
    have metadata changes affect the generated spec file (default: OFF)."""

    patch_foreign: bool = True
    """Automatically generate a patch that strips "foreign" dependencies from
    `Cargo.toml` (i.e. dependencies for non-Linux targets; default: ON)."""

    relative_license_paths: bool = False
    """Use relative paths instead of absolute paths for `%license` files in
    `%files` lists in spec files. This usually causes built RPM packages to
    contain two copies of these files (default: OFF)."""

    reuse_patch: bool = False
    """Attempt to re-apply existing `Cargo.toml` patch. Falls back to opening an
    editor if the `patch` flag is set (default: OFF)."""

    rpmautospec: bool | None = None
    """Override automatically detected use of rpmautospec in an existing spec
    file (default: fall back to autodetection)."""

    store_crate: bool = False
    """Copy the `.crate` file from the download cache into the output directory
    after downloading and processing it. This option has no effect when using
    local sources (default: OFF)."""

    validate_only: bool = False
    """Only attempt to load and validate a `rust2rpm.toml` configuration file
    and exit (default: OFF)."""

    vendor: VendorMode = VendorMode.OFF
    """Use vendored dependencies and / or create a tarball containing vendored
    dependencies (default: OFF)."""


@dataclass(frozen=True)
class CliArgs:
    """Parsed and processed command-line arguments."""

    pkgid: str | None
    """Project identifier consisting of name and version (both optional) in the
    format `<name>`, `<name>@<version>`, `@<version>`, or `None`."""

    output_directory: Path | None
    """Optional output directory (default: current working directory)."""

    path: Path | None
    """Optional path to a local `.crate` file, local `Cargo.toml` file, or a
    directory containing a `Cargo.toml` file for operating on local sources."""

    config_file: Path | None
    """Optional explicit path to a `rust2rpm.toml` configuration file
    (default: `rust2rpm.toml` in current working directory, if it exists)."""

    target: Target
    """Target to generate spec file for (default: based on current system)."""

    options: Options
    """Collection of other flags and flag-like command-line arguments."""

    @staticmethod
    def parse(args: list[str] | None = None) -> "CliArgs":
        """Parse a list of arguments into `CliArgs`."""
        parser = get_parser()
        parsed = parser.parse_args(args)

        output_directory = Path(parsed.output_directory) if parsed.output_directory else None
        path = Path(parsed.path) if parsed.path else None
        config_file = Path(parsed.config_file) if parsed.config_file else None
        target = Target(parsed.target) if parsed.target else Target(get_default_target())
        vendor = VendorMode(parsed.vendor) if parsed.vendor else VendorMode.OFF

        options = Options(
            auto_changelog_entry=parsed.auto_changelog_entry,
            compat=parsed.compat,
            existence_check=parsed.existence_check,
            ignore_missing_license_files=parsed.ignore_missing_license_files,
            offline=parsed.offline,
            patch=parsed.patch,
            patch_foreign=parsed.patch_foreign,
            relative_license_paths=parsed.relative_license_paths,
            reuse_patch=parsed.reuse_patch,
            rpmautospec=parsed.rpmautospec,
            store_crate=parsed.store_crate,
            validate_only=parsed.validate_only,
            vendor=vendor,
        )

        return CliArgs(
            pkgid=parsed.pkgid,
            output_directory=output_directory,
            path=path,
            config_file=config_file,
            target=target,
            options=options,
        )

    @property
    def name(self) -> str | None:
        """The optional `<name>` part of the `pkgid` argument."""
        if self.pkgid is None:
            return None

        if "@" in self.pkgid:
            name, _version = self.pkgid.split("@")
            return name if name != "" else None

        return self.pkgid

    @property
    def version(self) -> str | None:
        """The optional `<version>` part of the `pkgid` argument."""
        if self.pkgid is None:
            return None

        if "@" in self.pkgid:
            _name, version = self.pkgid.split("@")
            return version if version != "" else None

        return None
