"""Module containing functionality for interacting with the crate registry on <https://crates.io>."""

import contextlib
import os
from pathlib import Path
from urllib.parse import urljoin

import requests
import tqdm
from cargo2rpm.semver import Version

from rust2rpm import log

CRATES_IO_API_URL = "https://crates.io/api/v1/"
"""The root of all URLs of the crates.io API."""

if XDG_CACHE_HOME_STR := os.getenv("XDG_CACHE_HOME"):
    XDG_CACHE_HOME = Path(XDG_CACHE_HOME_STR)
else:
    XDG_CACHE_HOME = Path("~/.cache").expanduser()

CACHE_DIR = XDG_CACHE_HOME / "rust2rpm"
"""Location of the rust2rpm download cache.

This is where crate files that are downloaded from <https://crates.io> are
stored.
"""


class NoVersionsError(Exception):
    """Raised when querying crates.io yields no valid versions."""


def query_available_versions(
    crate: str,
    *,
    stable: bool = True,
) -> list[Version]:  # pragma nocover: requires internet access
    """Query the crates.io API for all available versions of a crate.

    Arguments:
        crate: Name of the crate in the registry.
        stable: Filter out pre-releases if set to `True`.

    Returns:
        List of available versions.

    """
    url = urljoin(CRATES_IO_API_URL, f"crates/{crate}/versions")
    req = requests.get(url, headers={"User-Agent": "rust2rpm"}, timeout=10)
    req.raise_for_status()
    versions = req.json()["versions"]

    parsed_versions = (Version.parse(x["num"]) for x in filter(lambda x: not x["yanked"], versions))

    if stable:
        return list(filter(lambda x: x.pre is None, parsed_versions))
    return list(parsed_versions)


def query_newest_version(crate: str) -> Version:  # pragma nocover: requires internet access
    """Query the crates.io API for the greatest available version of a crate.

    Arguments:
        crate: Name of the crate in the registry.

    Returns:
        The greatest available version.

    Raises:
        `NoVersionsError`: No version is available or is considered suitable.

    """
    url = urljoin(CRATES_IO_API_URL, f"crates/{crate}/versions")
    req = requests.get(url, headers={"User-Agent": "rust2rpm"}, timeout=10)
    req.raise_for_status()
    versions = req.json()["versions"]

    def is_stable(s: dict) -> bool:
        return Version.parse(s["num"]).pre is None

    def is_not_yanked(s: dict) -> bool:
        return not s["yanked"]

    # return the most recent, non-yanked stable version

    def is_not_yanked_and_stable(s: dict) -> bool:
        return is_stable(s) and is_not_yanked(s)

    not_yanked_and_stable = [*filter(is_not_yanked_and_stable, versions)]
    if len(not_yanked_and_stable) > 0:
        return Version.parse(not_yanked_and_stable[0]["num"])

    # there are no non-yanked stable versions:
    # fall back to the latest pre-release

    not_yanked = [*filter(is_not_yanked, versions)]
    if len(not_yanked) > 0:
        version = not_yanked[0]["num"]
        log.warn(f"No stable versions available. Falling back to the latest pre-release version {version!r}.")
        return Version.parse(version)

    # there are no non-yanked versions: fatal
    msg = f"No versions are available for crate {crate!r}."
    raise NoVersionsError(msg)


@contextlib.contextmanager
def remove_on_error(path: Path):  # pragma nocover: only used in download_crate
    """Context manager that removes the path on any error."""
    try:
        yield
    except:  # this is supposed to include ^C
        path.unlink()
        raise


def download_crate(name: str, version: Version, *, offline: bool) -> Path:  # pragma nocover: requires internet access
    """Download the specified crate version from crates.io.

    Arguments:
        name: Name of the crate in the registry.
        version: Version of the crate that is downloaded.
        offline: Skip download and only construct file path if set to `True`.

    Returns:
        Path to the file downloaded into the download cache directory.

    """
    CACHE_DIR.mkdir(parents=True, exist_ok=True)

    crate_file_name = f"{name}-{version}.crate"
    crate_file_path = CACHE_DIR / crate_file_name

    if not crate_file_path.is_file() and not offline:
        url = urljoin(CRATES_IO_API_URL, f"crates/{name}/{version}/download#")
        req = requests.get(url, stream=True, headers={"User-Agent": "rust2rpm"}, timeout=10)
        req.raise_for_status()
        total = int(req.headers["Content-Length"])

        with remove_on_error(crate_file_path), crate_file_path.open("wb") as f:
            for chunk in tqdm.tqdm(
                req.iter_content(),
                f"Downloading {crate_file_name}",
                total=total,
                unit="B",
                unit_scale=True,
            ):
                f.write(chunk)

    return crate_file_path
