"""Module containing functionality for interacting with the <https://src.fedoraproject.org> API."""

from urllib.parse import urljoin

import requests

DIST_GIT_API_URL = "https://src.fedoraproject.org/api/0/"


def get_package_info(package: str) -> dict | None:  # pragma nocover: requires internet access
    """Query src.fedoraproject.org API for information about a package.

    Arguments:
        package: The (source) name of the package under the `rpms/` namespace.

    Returns:
        Metadata in JSON format if the package exists in Fedora and is not
        retired in Rawhide, or `None` if the package is not present in Fedora.

    """
    # check if the dist-git repository exists

    url = urljoin(DIST_GIT_API_URL, f"rpms/{package}")
    req = requests.get(url, headers={"User-Agent": "rust2rpm"}, timeout=10)
    data = req.json()

    if "name" not in data:
        return None

    # check if there is a spec file on the rawhide branch

    full_url = data["full_url"]
    if not full_url.endswith("/"):
        # ensure the base URL has a trailing slash;
        # otherwise urljoin mangles the URL
        full_url += "/"

    spec_url = urljoin(full_url, f"blob/rawhide/f/{package}.spec")
    req = requests.head(spec_url, headers={"User-Agent": "rust2rpm"}, timeout=10)

    if not req.ok:
        return None

    return data
