"""Module containing various helper functionality for interacting with the filesystem."""

import os
import re
import shutil
import subprocess
from pathlib import Path

from rust2rpm import log
from rust2rpm.sysinfo import Target

DEFAULT_EDITOR = "nano"


class NoEditorError(Exception):
    """Exception raised when no suitable terminal editor can be found."""


def detect_editor() -> str:
    """Return the executable name of the default editor.

    Precedence:

    - Return the value of the `$VISUAL` environment variable if it is set.
    - Return the value of the `$EDITOR` environment variable if it is set.
    - Return the default editor (`DEFAULT_EDITOR`) if neither are set.

    Returns:
        Name of the editor executable, which is expected to be available in `$PATH`.

    Raises:
        `NoEditorError`: The current terminal is too dumb to handle interactive use.

    """
    terminal = os.getenv("TERM")
    terminal_is_dumb = not terminal or terminal == "dumb"
    editor = None
    if not terminal_is_dumb:
        editor = os.getenv("VISUAL")
    if not editor:
        editor = os.getenv("EDITOR")
    if not editor:
        if terminal_is_dumb:
            msg = "Terminal is dumb, but $EDITOR unset"
            raise NoEditorError(msg)
        editor = DEFAULT_EDITOR
    return editor


def detect_packager() -> str | None:
    """Return an identifier for the current user in the `name <email>` format.

    Precedence:

    - `None` if the `RUST2RPM_NO_DETECT_PACKAGER` environment variable is set.
    - The value of the `RUST2RPM_PACKAGER` environment variable if it is set.
    - The string printed by the `rpmdev-packager` tool if it is available.
    - A user identifier based on the current user's `git` configuration.
    - `None` if none of the previous were available.

    Returns:
        Identifier for the current user for use in RPM changelog entries.

    """
    # If we're forcing the fallback...
    if os.getenv("RUST2RPM_NO_DETECT_PACKAGER"):
        return None

    # If we're supplying packager identity through an environment variable...
    if packager := os.getenv("RUST2RPM_PACKAGER"):
        return packager

    # If we're detecting packager identity through rpmdev-packager...
    if rpmdev_packager := shutil.which("rpmdev-packager"):
        return subprocess.check_output(rpmdev_packager, text=True).strip()  # noqa: S603

    # If we're detecting packager identity through git configuration...
    if git := shutil.which("git"):
        name = subprocess.check_output([git, "config", "user.name"], text=True).strip()  # noqa: S603
        email = subprocess.check_output([git, "config", "user.email"], text=True).strip()  # noqa: S603
        return f"{name} <{email}>"

    return None


def guess_crate_name(root: Path | None = None) -> str | None:
    """Return guessed crate name based on directory, spec file, and spec file contents.

    Precedence:

    - If exactly one file matching `*.spec` is present in the root directory,
        check contents for the presence of a `%crate` macro and return its value.
        This supports spec files for *"compat"* packages, where spec file names
        cannot be uniquely parsed into (name, version).
    - If more than one file matching `*.spec` are present in the root directory,
        return `None` and require specifying the crate name explicitly in CLI
        arguments.
    - If no files matching `*.spec` are present in the root directory, return
        a guess based on the name of the root directory.
    - If none of the previous heuristics matched, return `None` and require
        specifying the crate name explicitly in CLI arguments.

    Arguments:
        root: Override root directory for heuristics. Defaults to the current directory.

    Returns:
        Name of the current crate based on heuristics.

    """
    path = root or Path()
    specs = [*path.glob("*.spec")]

    if len(specs) > 1:
        log.error(
            "Found multiple spec files in the current working directory; "
            "unable to determine crate name automatically.",
        )
        return None

    if len(specs) == 1:
        crate = None
        spec = specs[0]

        with spec.open() as file:
            for line in file:
                if m := re.match(r"^%(?:global|define)\s+crate\s+(\S+)\s+", line):
                    if crate:
                        log.error(
                            f"Found multiple definitions of the '%crate' macro in {spec!s}; "
                            "unable to determine crate name automatically.",
                        )
                        return None
                    crate = m.group(1)
                    if "%" in crate:
                        log.error("The value of the %crate macro appears to contain other macros and cannot be parsed.")
                        return None

        if crate:
            log.success(f"Found valid spec file {spec!s} for the {crate!r} crate.")
        else:
            log.error(f"Invalid spec file {spec!s}; unable to determine crate name automatically.")
        return crate

    if m := re.match("^rust-([a-z+0-9_-]+)$", path.absolute().name):
        crate = m.group(1)
        log.info(f"Continuing with crate name {crate!r} based on the current working directory.")
        return crate

    return None


def detect_rpmautospec(target: Target, spec_file: Path) -> bool:
    """Return guess whether %autorelease+%autochangelog should be used.

    Precedence:

    - False if the target is not Fedora or EPEL 8
    - If a spec file already exists, return whether it is using rpmautospec.
    - If no spec file exists yet, return True.

    Arguments:
        target: `Target` for the spec file generator.
        spec_file: Path to a spec file that might or might not exist.

    Returns:
        Boolean flag that determines whether rpmautospec will be used.

    """
    # We default to on only for selected distros for now…
    if target not in {Target.FEDORA, Target.EPEL8}:
        return False

    try:
        text = spec_file.read_text()
    except FileNotFoundError:
        # A new file, let's try the new thing.
        return True

    # We are redoing an existing spec file. Figure out if it had
    # %autochangelog enabled before.
    autochangelog_re = r"^\s*%(?:autochangelog|\{\??autochangelog\})\b"
    return any(re.match(autochangelog_re, line) for line in text.splitlines())
